from typing import Any, Generator, Sequence
from collections.abc import Sequence as SequenceType


def serve(items: Sequence[Any], serving_size: int) -> Generator[Sequence[Any], None, None]:
    """
    Split a sequence of items into a number of smaller sequences, with max serving_size items in each.

    :param items: The original sequence of items.
    :param serving_size: The max number of items in each outputted subsequence.
    :return: Generator yielding sliced subsequences.
    """

    if not isinstance(items, SequenceType):
        raise NotImplementedError('Dollop sequence.serve only supports objects of Sequence type.')

    for i in range(0, len(items), serving_size):
        yield items[i: i + serving_size]



import pytest

from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from dollop.file import serve


def helper_create_and_serve_file_obj(content, file_obj_type, mode, serving_size):

    with TemporaryDirectory() as td:

        file_path = f'{td}/content.txt'
        with open(file_path, 'w') as f:
            f.write(content)

        if file_obj_type == 'string':
            file_obj = file_path
        elif file_obj_type == 'pathlib':
            file_obj = Path(file_path)
        elif file_obj_type == 'handle':
            file_obj = open(file_path, 'r')
        elif file_obj_type == 'handle_binary':
            file_obj = open(file_path, 'rb')
        else:
            raise ValueError('Invalid file_obj_type.')

        dollops = [*serve(file_obj, serving_size=serving_size, mode=mode)]
        return dollops


def helper_test_served_file_obj(content, dollops, mode, serving_size,
                                expected_n_full_servings, expected_remainder, binary):

    subdollops = [d.splitlines() for d in dollops] if mode == 'lines' else dollops

    if expected_remainder == 0:
        assert len(dollops) == expected_n_full_servings
        assert all([len(d) == serving_size for d in subdollops])

    else:
        assert len(dollops) == expected_n_full_servings + 1
        assert all([len(d) == serving_size for d in subdollops[:-1]])
        assert len(subdollops[-1]) == expected_remainder

    if binary:
        assert all([type(d) == bytes for d in dollops])
        undolloped = ''.join([d.decode() for d in dollops])
    else:
        assert all([type(d) == str for d in dollops])
        undolloped = ''.join(dollops)

    assert undolloped.splitlines() == content.splitlines()


@pytest.mark.parametrize('mode', ('characters', 'lines'))
@pytest.mark.parametrize('file_obj_type', ('string', 'pathlib', 'handle', 'handle_binary'))
def test_serve_empty_file(mode, file_obj_type):
    content = ''
    dollops = helper_create_and_serve_file_obj(content, file_obj_type, mode, serving_size=10)
    assert len(dollops) == 0


@pytest.mark.parametrize(
    'mode,serving_size,expected_n_full_servings,expected_remainder',
    (
            ('lines', 10, 1, 0),  # everything in one serving
            ('lines', 5, 2, 0),  # split, no remainder
            ('lines', 3, 3, 1),  # split, remainder
            ('lines', 20, 0, 10),  # no full servings, remainder
            ('characters', 133, 1, 0),  # everything in one serving
            ('characters', 19, 7, 0),  # split, no remainder
            ('characters', 20, 6, 13),  # split, remainder
            ('characters', 200, 0, 133),  # no full servings, remainder
    )
)
@pytest.mark.parametrize('file_obj_type', ('string', 'pathlib', 'handle', 'handle_binary'))
def test_serve_nonempty_file_poem(mode, file_obj_type,
                                  serving_size, expected_n_full_servings, expected_remainder):
    content = '\"Fog\"\n' + \
              'by Carl Sandburg\n' + \
              '\n' + \
              'The fog comes\n' + \
              'on little cat feet.\n' + \
              '\n' + \
              'It sits looking\n' + \
              'over harbor and city\n' + \
              'on silent haunches\n' + \
              'and then moves on.'

    dollops = helper_create_and_serve_file_obj(content, file_obj_type, mode, serving_size)
    helper_test_served_file_obj(content, dollops, mode, serving_size,
                                expected_n_full_servings, expected_remainder, binary=(file_obj_type=='handle_binary'))


@pytest.mark.parametrize(
    'mode,serving_size,expected_n_full_servings,expected_remainder',
    (
            ('lines', 10, 1, 0),  # everything in one serving
            ('lines', 5, 2, 0),  # split, no remainder
            ('lines', 3, 3, 1),  # split, remainder
            ('lines', 20, 0, 10),  # no full servings, remainder
            ('characters', 40, 1, 0),  # everything in one serving
            ('characters', 20, 2, 0),  # split, no remainder
            ('characters', 17, 2, 6),  # split, remainder
            ('characters', 50, 0, 40),  # no full servings, remainder
    )
)
@pytest.mark.parametrize('file_obj_type', ('string', 'pathlib', 'handle', 'handle_binary'))
def test_serve_nonempty_file_blank_space(mode, file_obj_type,
                                         serving_size, expected_n_full_servings, expected_remainder):
    content = '\n' + \
              '\n' + \
              '\n' + \
              'nice to\n' + \
              '\n' + \
              'meet you\n' + \
              '\n' + \
              '\n' + \
              '\n' + \
              'where you been?\n'

    dollops = helper_create_and_serve_file_obj(content, file_obj_type, mode, serving_size)
    helper_test_served_file_obj(content, dollops, mode, serving_size,
                                expected_n_full_servings, expected_remainder, binary=(file_obj_type=='handle_binary'))


def test_unknown_mode_raises_error():
    with pytest.raises(ValueError):
        _ = [*serve('file.txt', serving_size=1, mode='words')]


def test_write_handle_raises_error():
    with NamedTemporaryFile() as tf:
        with pytest.raises(ValueError):
            _ = [*serve(open(tf.name, 'w'), serving_size=1)]


def test_nonexistent_file_raises_error():
    with TemporaryDirectory() as td:
        with pytest.raises(FileNotFoundError):
            _ = [*serve(f'{td}/file.txt', serving_size=1)]



from typing import Generator, TYPE_CHECKING

if TYPE_CHECKING:
    import pandas # Keep this here for string-based type-checking


def serve(pandas_obj: "pandas.DataFrame | pandas.Series", serving_size: int) -> \
        Generator["pandas.DataFrame | pandas.Series", None, None]:
    """
    Read a Pandas object small chunks at a time.

    :param pandas_obj: The Pandas object.
    :param serving_size: The max number of items in each serving.
    :return: Generator yielding sliced pandas objects.
    """

    cls = pandas_obj.__class__
    is_dataframe = cls.__module__.startswith("pandas.core.frame") and cls.__name__ == "DataFrame"
    is_series = cls.__module__.startswith("pandas.core.series") and cls.__name__ == "Series"

    if not (is_dataframe or is_series):
        raise TypeError(f"Dollop pandas.serve only supports pandas DataFrame or Series")

    for i in range(0, len(pandas_obj), serving_size):
        yield pandas_obj.iloc[i: i + serving_size]




import pytest

from dollop.sequence import serve


@pytest.fixture
def sequence_creators():

    def create_list(n):
        return list(range(n))

    def create_tuple(n):
        return tuple(range(n))

    def create_string(n):
        return 'a' * n

    def create_range(n):
        return range(n)

    def create_bytes(n):
        return bytes(n)

    def create_bytearray(n):
        return bytearray(n)

    return {
        'list': create_list,
        'tuple': create_tuple,
        'str': create_string,
        'range': create_range,
        'bytes': create_bytes,
        'bytearray': create_bytearray,
    }


@pytest.fixture
def non_sequences():
    return {
        'int': 10,
        'float': 10.1,
        'none': None,
    }


@pytest.mark.parametrize('sequence_type', ('list', 'tuple', 'str', 'range', 'bytes', 'bytearray'))
@pytest.mark.parametrize(
    'n_items,serving_size,expected_n_full_servings,expected_remainder',
    (
            (0, 10, 0, 0),  # empty
            (10, 10, 1, 0),  # everything in one serving
            (10, 5, 2, 0),  # split, no remainder
            (10, 3, 3, 1),  # split, remainder
            (10, 15, 0, 10),  # no full servings, remainder
    )
)
def test_sequence_works(n_items, serving_size, expected_n_full_servings, expected_remainder,
                        sequence_type, sequence_creators):
    create_sequence = sequence_creators[sequence_type]

    items = create_sequence(n=n_items)
    dollops = [*serve(items, serving_size=serving_size)]

    if expected_remainder == 0:
        assert len(dollops) == expected_n_full_servings
        assert all([len(d) == serving_size for d in dollops])
    else:
        assert len(dollops) == expected_n_full_servings + 1
        assert all([len(d) == serving_size for d in dollops[:-1]])
        assert len(dollops[-1]) == expected_remainder

    expected_items = [x for c in dollops for x in c]
    assert all(x == y for x, y in zip(expected_items, items))

    assert all([type(serving) == type(items) for serving in dollops])


@pytest.mark.parametrize('non_sequence_type', ('int', 'float', 'none'))
def test_non_sequence_raises_error(non_sequence_type, non_sequences):
    non_sequence = non_sequences[non_sequence_type]
    with pytest.raises(NotImplementedError):
        _ = [*serve(non_sequence, serving_size=10)]



import numpy as np
import pytest

import torch
from dollop.torch import serve

sizes = (
    (1,),
    (10,),
    (100,),
    (100, 100, 10),
    (101, 99, 67),
    (101, 1, 67),
    (101, 1, 67, 902),
    (20, 20, 20, 20),
    (1, 1, 1, 1, 1, 1),
)

arrays = [torch.rand(size) for size in sizes]


@pytest.mark.parametrize("tensor", arrays)
@pytest.mark.parametrize("dim", list(range(5)))
@pytest.mark.parametrize("serving_size", [1, 2, 3, 5, 10, 21, 25, 27, 100, 200])
def test_serve_torch(tensor, dim, serving_size):
    if tensor.ndim <= dim:
        with pytest.raises(ValueError):
            _ = [*serve(tensor, serving_size=serving_size, dim=dim)]
        return

    size = tensor.shape[dim]
    other_dims = [i for i in range(tensor.ndim) if i != dim]
    other_shapes = [tensor.shape[i] for i in other_dims]

    dollops = [*serve(tensor, serving_size=serving_size, dim=dim)]

    expected_full = size // serving_size
    remainder = size % serving_size

    if remainder == 0:
        assert len(dollops) == expected_full
        assert all(d.shape[dim] == serving_size for d in dollops)
    else:
        assert len(dollops) == expected_full + 1
        assert all(d.shape[dim] == serving_size for d in dollops[:-1])
        assert dollops[-1].shape[dim] == remainder

    for d in dollops:
        for i, shape in zip(other_dims, other_shapes):
            assert d.shape[i] == shape, f"Mismatch on dim {i}: expected {shape}, got {d.shape[i]}"

    recon = torch.cat(dollops, dim=dim)
    np.testing.assert_array_equal(recon.numpy(), tensor.numpy())
    assert all(isinstance(d, torch.Tensor) for d in dollops)


def test_serve_torch_0d_tensor_raises():
    scalar = torch.tensor(42)
    with pytest.raises(ValueError):
        _ = [*serve(scalar, serving_size=1)]


@pytest.mark.parametrize("bad_input", [
    [1, 2, 3],
    (1, 2, 3),
    10,
    10.5,
    None,
])
def test_serve_torch_non_tensor_obj_raises_error(bad_input):
    with pytest.raises(TypeError):
        _ = [*serve(bad_input, serving_size=5)]



import io
import numpy as np
import pandas as pd
import pathlib
import pytest
import torch

from unittest.mock import Mock, patch

from dollop.auto import serve as serve


@pytest.fixture
def sequence_types():
    return {
        'list': list,
        'tuple': tuple,
        'str': str,
        'range': range,
        'bytes': bytes,
        'bytearray': bytearray,
    }


@pytest.fixture
def file_types():
    return {
        'handle': io.IOBase,
        'pathlib': pathlib.Path,
    }


@pytest.fixture
def pandas_types():
    return {
        'series': pd.Series,
        'df': pd.DataFrame,
    }


@pytest.fixture
def torch_types():
    return {
        'tensor': torch.Tensor,
    }


@pytest.fixture
def numpy_types():
    return {
        'array': np.ndarray,
    }


@pytest.fixture
def other_types():
    return {
        'int': int,
        'float': float,
        'none': None,
    }


@pytest.mark.parametrize('sequence_type', ('list', 'tuple', 'str', 'range', 'bytes', 'bytearray'))
def test_serve_auto_sequence_type(sequence_type, sequence_types):
    mock_obj = Mock(spec=sequence_types[sequence_type])
    with patch('dollop.auto.serve_sequence') as serve_sequence:
        serve_sequence.return_value = iter([None])
        _ = [*serve(mock_obj, serving_size=10)]
        serve_sequence.assert_called_once()


@pytest.mark.parametrize('file_type', ('handle', 'pathlib'))
def test_serve_auto_file_type(file_type, file_types):
    mock_obj = Mock(spec=file_types[file_type])
    with patch('dollop.auto.serve_file') as serve_file:
        serve_file.return_value = iter([None])
        _ = [*serve(mock_obj, serving_size=10)]
        serve_file.assert_called_once()


@pytest.mark.parametrize('pandas_type', ('series', 'df'))
def test_serve_auto_pandas_type(pandas_type, pandas_types):
    mock_obj = Mock(spec=pandas_types[pandas_type])
    with patch('dollop.auto.serve_pandas') as serve_pandas:
        serve_pandas.return_value = iter([None])
        _ = [*serve(mock_obj, serving_size=10)]
        serve_pandas.assert_called_once()


@pytest.mark.parametrize('numpy_type', ('array',))
def test_serve_auto_numpy_types(numpy_type, numpy_types):
    mock_obj = Mock(spec=numpy_types[numpy_type])
    with patch('dollop.auto.serve_numpy') as serve_numpy:
        serve_numpy.return_value = iter([None])
        _ = [*serve(mock_obj, serving_size=10)]
        serve_numpy.assert_called_once()


@pytest.mark.parametrize('torch_type', ('tensor',))
def test_serve_auto_torch_type(torch_type, torch_types):
    mock_obj = Mock(spec=torch_types[torch_type])
    with patch('dollop.auto.serve_torch') as serve_torch:
        serve_torch.return_value = iter([None])
        _ = [*serve(mock_obj, serving_size=10)]
        serve_torch.assert_called_once()


@pytest.mark.parametrize('other_type', ('int', 'float', 'none'))
def test_serve_auto_other_type(other_type, other_types):
    mock_obj = Mock(spec=other_types[other_type])
    with patch('dollop.auto.serve_sequence') as serve_sequence:
        with pytest.raises(NotImplementedError):
            _ = [*serve(mock_obj, serving_size=10)]



import io
import pathlib

from typing import Union, Generator


def serve(file: Union[str, pathlib.Path, io.IOBase], serving_size: int, mode='lines') \
                -> Generator[str, None, None]:
    """
    Read a file small dollops at a time.

    :param file: The path or file handle.
    :param serving_size: The max number of lines in each serving.
    :param mode: Either 'lines' or 'chars'/'characters' to read line-by-line or character-by-character.
    :return:
    """

    if mode == 'lines':
        def read_next(h):
            try:
                return next(h)
            except StopIteration:
                return ''

    elif mode in ('characters', 'chars'):
        def read_next(h):
            return h.read(1)

    else:
        raise ValueError('Invalid mode for dollop.file.serve')

    def _serve_handle(h):
        binary = 'b' in h.mode

        finished = False
        while not finished:

            dollop = b'' if binary else ''
            for _ in range(serving_size):
                token = read_next(h)
                if token:
                    dollop += token
                else:
                    finished = True
                    break

            # Check for edge case where there's nothing left.
            if dollop:
                yield dollop

    if isinstance(file, io.IOBase):
        if 'r' not in file.mode:
            raise ValueError('Can only dollop file handles with a read attribute.')
        yield from _serve_handle(file)

    elif isinstance(file, (str, pathlib.Path)):
        with open(file, 'r') as f:
            yield from _serve_handle(f)



from typing import Generator, TYPE_CHECKING

if TYPE_CHECKING:
    import torch # Keep this here for string-based type-checking


def serve(tensor: "torch.Tensor", serving_size: int, dim: int = 0) -> Generator["torch.Tensor", None, None]:
    """
    Read a PyTorch tensor small chunks at a time.

    :param tensor: The tensor object.
    :param serving_size: The max number of items in each serving.
    :param dim: The dimension along which to slice. Default is 0.
    :return: Generator yielding tensor slices.
    """
    if not (tensor.__class__.__module__.startswith("torch") and tensor.__class__.__name__ == "Tensor"):
        raise TypeError('Dollop torch.serve only supports PyTorch Tensor types.')

    if tensor.ndim == 0:
        raise ValueError("Cannot slice a zero-dimensional (scalar) tensor.")

    if dim < 0 or dim >= tensor.ndim:
        raise ValueError(f"Invalid dim={dim}; tensor has {tensor.ndim} dimensions.")

    # Make a view of the array with the chosen dim at the front.
    axes = list(range(tensor.ndim))
    axes[0], axes[dim] = axes[dim], axes[0]
    tensor_permuted = tensor.permute(*axes)

    # Get inverse permutation
    axes_inv = [axes.index(i) for i in range(tensor.ndim)]

    for i in range(0, tensor_permuted.shape[0], serving_size):
        dollop = tensor_permuted[i: i + serving_size]
        yield dollop.permute(*axes_inv)



import io
import pathlib

from typing import Any, Generator
from collections.abc import Sequence as SequenceType

from .sequence import serve as serve_sequence
from .file import serve as serve_file
from .pandas import serve as serve_pandas
from .numpy import serve as serve_numpy
from .torch import serve as serve_torch


def serve(obj: Any, serving_size: int, **kwargs) -> Generator[Any, None, None]:
    """
    Split an iterable of items into a number of smaller iterables, with max serving_size items
    in each. Detect the type automatically and decide how to proceed.

    :param obj: The object that we wish to serve.
    :param serving_size: The max number of items in each outputted subiterable.
    :param kwargs: Any additional arguments to pass to the type-specific function.
    :return:
    """

    if isinstance(obj, SequenceType):
        yield from serve_sequence(obj, serving_size=serving_size)

    elif isinstance(obj, (io.IOBase, pathlib.Path)):
        yield from serve_file(obj, serving_size=serving_size, **kwargs)

    elif obj.__class__.__module__.startswith("pandas"):
        yield from serve_pandas(obj, serving_size=serving_size)

    elif obj.__class__.__module__.startswith("numpy"):
        yield from serve_numpy(obj, serving_size=serving_size, **kwargs)

    elif obj.__class__.__module__.startswith("torch"):
        yield from serve_torch(obj, serving_size=serving_size, **kwargs)

    else:
        raise NotImplementedError(f'Object of type {type(obj)} is not dollopable.')



import pandas as pd
import pytest

from dollop.pandas import serve


@pytest.fixture
def pandas_obj_creators():

    def create_series(n):
        data = [*range(n)]
        series = pd.Series(data)
        series.index = [f'column_{i}' for i in range(n)]
        return series

    def create_dataframe(n):
        n_columns = 10
        data = [[10 ** j + i for j in range(n)] for i in range(n_columns)]
        df = pd.DataFrame(data).T
        df.columns = [f'column_{i}' for i in range(n_columns)]
        return df

    return {
        'series': create_series,
        'df': create_dataframe,
    }


@pytest.fixture
def non_pandas_objs():
    return {
        'list': [1, 2, 3],
        'tuple': (1, 2, 3),
        'int': 10,
        'float': 10.1,
        'none': None,
    }


@pytest.mark.parametrize('pandas_obj_type', ('series', 'df'))
@pytest.mark.parametrize(
    'n_items,serving_size,expected_n_full_servings,expected_remainder',
    (
            (0, 10, 0, 0),  # empty
            (10, 10, 1, 0),  # everything in one serving
            (10, 5, 2, 0),  # split, no remainder
            (10, 3, 3, 1),  # split, remainder
            (10, 15, 0, 10),  # no full servings, remainder
    )
)
def test_pandas_obj_works(n_items, serving_size, expected_n_full_servings, expected_remainder,
                      pandas_obj_type, pandas_obj_creators):

    create_pandas_obj = pandas_obj_creators[pandas_obj_type]

    pandas_obj = create_pandas_obj(n=n_items)
    dollops = [*serve(pandas_obj, serving_size=serving_size)]

    if expected_remainder == 0:
        assert len(dollops) == expected_n_full_servings
        assert all([len(d) == serving_size for d in dollops])
    else:
        assert len(dollops) == expected_n_full_servings + 1
        assert all([len(d) == serving_size for d in dollops[:-1]])
        assert len(dollops[-1]) == expected_remainder

    if n_items != 0:
        assert (pd.concat(dollops) == pandas_obj).all().all()
        assert all([type(serving) == type(pandas_obj) for serving in dollops])


@pytest.mark.parametrize('non_sequence_type', ('list', 'tuple', 'int', 'float', 'none'))
def test_non_pandas_obj_raises_error(non_sequence_type, non_pandas_objs):
    non_sequence = non_pandas_objs[non_sequence_type]
    with pytest.raises(TypeError):
        _ = [*serve(non_sequence, serving_size=10)]



from typing import Generator, TYPE_CHECKING

if TYPE_CHECKING:
    import numpy # Keep this here for string-based type-checking


def serve(array: "numpy.ndarray", serving_size: int, dim: int = 0) -> Generator["numpy.ndarray", None, None]:
    """
    Read a NumPy array small chunks at a time.

    :param array: The NumPy array to slice.
    :param serving_size: The number of items per slice.
    :param dim: The dimension along which to slice. Default is 0.
    :return: Generator yielding array slices.
    """
    if not (array.__class__.__module__.startswith("numpy") and array.__class__.__name__ == "ndarray"):
        raise TypeError('Dollop numpy.serve only supports NumPy ndarray types.')

    if array.ndim == 0:
        raise ValueError("Cannot slice a zero-dimensional (scalar) array.")

    if dim < 0 or dim >= array.ndim:
        raise ValueError(f"Specified dim ({dim}) must be smaller than the number of array dimensions ({array.ndim})")

    # Make a view of the array with the chosen dim at the front.
    axes = list(range(array.ndim))
    axes[0], axes[dim] = axes[dim], axes[0]
    array_permutated = array.transpose(axes)

    # Get inverse permutation
    axes_inv = [axes.index(i) for i in range(array.ndim)]

    for i in range(0, len(array_permutated), serving_size):
        dollop = array_permutated[i: i + serving_size]
        yield dollop.transpose(axes_inv)
